'   sdfat:  read access to a FAT16 file system in the first partition
'   of a secure digital memory card.
'
'   For the electrical connection expected, see the sdspi source file.
'
'   General usage:
'      mount(basepin)
'      stat := popenr(name)
'      count := pread(buf, count)
'      char := pgetc
'
'   Only mounts the first partition.  Only supports FAT16, and only up
'   to 2GB volumes.
'
'   Currently only supports one file open at a time, but this is very
'   easy to extend.  Does not yet support seeking or writing, but again
'   these are straightforward to add.
'
'   All failures are returned via abort with a pointer to a string that
'   indicates why the failure occurred.
'
con
   sectorsize = 512   ' do not change these constants
   sectorshift = 9
   dirsize = 32
   dirshift = 5
   dirpersector = sectorsize / dirsize
   dpsshift = sectorshift - dirshift
obj
   sdspi: "sdspi"
var
   ' volume entries; one copy per card
   long rootdir, rootentries, clustershift, fat1, totclusters
   ' per-file entries; static for now
   byte buf[512], padname[11]
   long fclust, filesize, floc, frem, bufat, bufend
pri brword(off)
'
'   Read a byte-reversed word from the buffer.
'
   return buf[off] + (buf[off+1]<<8)
pri brlong(off)
'
'   Read a byte-reversed long from the buffer.
'
   return brword(off) + (brword(off+2)<<16)
pub mount(basepin) | startblock, size, sectorspercluster, sectors, reserved, sectorsperfat
'
'   Mount a secure digital card.  Initialize the card,
'   find the first partition, find the relevant
'   FAT16 parameters and the location of the root
'   directory.  If all this works, return success.
'
   sdspi.start(basepin)
   sdspi.readblock(0, @buf)    ' read the partition table
   startblock := brlong($1c6)  ' start from first partition entry
   size := brlong($1ca)
   if size < 8
      abort string("error reading partition table")
   sdspi.readblock(startblock, @buf)  ' read the MBR
   if brword($1fe) <> $aa55           ' check for a valid signature
      abort string("bad mbr signature")
   if (brlong($36) <> "F" + ("A"<<8) + ("T"<<16) + ("1"<<24)) or buf[$3a] <> "6"
      abort string("not fat16")       ' if it's not FAT16, bail
   if (brword($0b) <> sectorsize)     ' sector size must be 512
      abort string("bad sectorsize")
   sectorspercluster := buf[$0d]
   if sectorspercluster & (sectorspercluster-1) <> 0
      abort string("bad sectors per cluster")
   clustershift := 0
   repeat while sectorspercluster > 1
      clustershift++
      sectorspercluster >>= 1
   if buf[$10] <> 2
      abort string("not 2 fat tables?")
   rootentries := brword($11)
   sectors := brword($13)
   if sectors == 0
      sectors := brlong($20)
   reserved := brword($0e)
   sectorsperfat := brword($16)
   fat1 := startblock + reserved
   rootdir := fat1 + (sectorsperfat << 1)
   totclusters := (sectors - reserved - (sectorsperfat << 1)) >> clustershift
pub popenr(name) | i, j, k, at, diff
'
'   Try to find the directory entry for name (which is a pointer
'   to a null-terminated string).  If successful, returns 0.
'   If the file is not found, returns -1.  Any error causes an
'   abort.
'
   i := 0
   j := 0
   repeat while i < 8 and byte[name][i] and byte[name][i] <> "."
      padname[j++] := byte[name][i++]
   repeat while j < 8
      padname[j++] := " "
   repeat while byte[name][i] and byte[name][i] <> "."
      i++
   if byte[name][i] == "."
      i++
   repeat while j < 11 and byte[name][i]
      padname[j++] := byte[name][i++]
   repeat while j < 11
      padname[j++] := " "
   i := 0
   repeat while i < rootentries >> dpsshift
      sdspi.readblock(rootdir + i, @buf)
      j := 0
      repeat j from 0 to sectorsize-dirsize step dirsize
         at := @buf + j
         k := byte[at][$0b]
         if k == 0 or k & $18
            next
         k := 0
         repeat while k < 11
            diff := padname[k] ^ byte[at][k]
            if diff <> 0
               if diff <> 32
                  quit
               diff := padname | 32
               if diff < "a" or diff > "z"
                  quit
            k++
         if k == 11
            fclust := brword(j+$1a)
            filesize := brlong(j+$1c)
            floc := 0
            frem := sectorsize << clustershift
            if frem + floc > filesize
               frem := filesize + floc
            bufat := 0
            bufend := 0
            return 0   
      i++
   return -1
pub pfillbuf | fatloc, fblock, r
'
'   Read the next block of the file (might be partial)
'
   if floc => filesize
      return -1
   if frem == 0
      fatloc := fat1 + (fclust >> (sectorshift-1))
      sdspi.readblock(fatloc, @buf)
      fclust := brword((fclust & constant((sectorsize >> 1) - 1)) << 1)
      if fclust < 2 or fclust => totclusters
         abort string("bad cluster value")
      frem := sectorsize << clustershift
      if frem + floc > filesize
         frem := filesize - floc
   fblock := ((fclust - 2) << clustershift) + rootdir + (rootentries >> dpsshift)
   fblock += (floc >> sectorshift) & ((1 << clustershift) - 1)
   sdspi.readblock(fblock, @buf)
   r := sectorsize
   if floc + r => filesize
      r := filesize - floc
   floc += r
   frem -= r
   bufat := 0
   bufend := r
   return r
pub pread(ubuf, togo) | r, t, more
'
'   Read togo bytes from the file into the user buffer ubuf.
'   Value returned is actual number of bytes read or
'   -1 if we have hit the end of file.
'
   r := 0
   repeat while togo > 0
      if bufat => bufend
         t := pfillbuf
         if t =< 0
            if r > 0
               return r
             return t
      more := bufend - bufat
      if more > togo
         more := togo
      bytemove(ubuf, buf+bufat, more)
      bufat += more
      r += more
      ubuf += more
      togo -= more
   return r
pub pgetc : t
'
'   Read and return a single character from the file.
'   Returns -1 on an error.
'
   if bufat => bufend
      t := pfillbuf
      if t < 0
         return t
      if t == 0
         return -1
   return buf[bufat++]